import os
import sys
import argparse
import getpass
import datetime
import shutil
import subprocess

import lxml.etree


SILECS_VERSION = "2.3.0"
SILECSDESIGN = "silecsdesign"
SILECSDEPLOY = "silecsdeploy"
DESIGN = "design"
DEPLOY = "deploy"
USER = getpass.getuser()

# FESA specific variables
FESA_VERSION="7.4.0"
FESA_BASE=f"/opt/fesa/fesa-fwk/{FESA_VERSION}"
FESA_XSD=f"{FESA_BASE}/fesa-model-gsi/xml/deployment/deployment-gsi.xsd"
FESA_GSI_TEMPLATE=f"{FESA_BASE}/fesa-model-gsi/xml/design/templates/GSIClassTemplate.xml"


# silecs specific variables
try:
    SILECS_BASE=(os.environ['SILECS_BASE'])
except KeyError:
    SILECS_BASE=f"/common/usr/cscofe/silecs/{SILECS_VERSION}"

SILECS_BASE = f".."
DESIGN_SHEMA_PATH = f"{SILECS_BASE}/silecs-model/src/xml/DesignSchema.xsd"
DEPLOY_SHEMA_PATH = f"{SILECS_BASE}/silecs-model/src/xml/DeploySchema.xsd"

SILECS_CODEGEN_BASE=f"{SILECS_BASE}/silecs-codegen/xml/"
SILECS_CODEGEN_MIGRATION=f"{SILECS_BASE}/silecs-codegen/xml/migration"
# TODO: Check for concrete fesa version
# See here for details: https://gitlab.com/al.schwinn/silecs-cli/-/issues/4
SILECS_CODEGEN_FESA=f"{SILECS_BASE}/silecs-codegen/xml/fesa/fesa_7_3_0/"

# import files for generating sources
sys.path.append(SILECS_CODEGEN_BASE)
sys.path.append(SILECS_CODEGEN_MIGRATION)
sys.path.append(SILECS_CODEGEN_FESA)

try:
    import genparam
    import genplcsrc
    import genduwrapper
    import fillFESADeployUnit
    import generateFesaDesign
    import generateSourceCode
except ModuleNotFoundError as e:
    print(f"Error during importing module: {e}")


print(f"using sileces base path: {SILECS_BASE}")


class SilecsArgumentParser(argparse.ArgumentParser):
    """Own argument parser"""
    def __init__(self, *args, **kwargs):
        argparse.ArgumentParser.__init__(self, *args, **kwargs)

    def show_examples(self):
        examples = "\nExamples: \n\n" \
                    "\t To create a new file 'MyClass.silecsdesign' for the FESA class. \n" \
                    "\t\t silces -c ~/workspace/MyClass/src/MyClass.design \n\n" \
                    "\t To validate 'MyClass.silecsdesign'. \n" \
                    "\t\t silces -v ~/workspace/MyClass/src/MyClass.silecsdesign \n\n" \
                    "\t To add missing properties and value-items specified in 'MyClass.silecsdesign' into the FESA document 'MyClass.design'.\n" \
                    "\t This will as well ate C++ code.\n" \
                    "\t\t silces -g ~/workspace/MyClass/src/MyClass.silecsdesign\n\n" \
                    "\nCheck the Silecs Wiki for more info: https://www-acc.gsi.de/wiki/Frontend/SILECS\n" \

        print(examples)

    def print_help(self, file=None):
        super().print_help(file)
        self.show_examples()

###### Helper functions #####

def get_extension(filepath):
    return filepath.split('.')[-1]


def get_project_name(filepath):
    """Get project name
     eg. workspace/TestClass/src/TestClass.silecsdesign -> TestClass
     """
    return os.path.splitext(os.path.basename(filepath))[0]


def get_project_dir(filepath):
    """Get project dir
     eg. workspace/TestClass/src/TestClass.silecsdesign -> workspace/TestClass/src"""
    return os.path.dirname(filepath)


def get_project_path(filepath):
    """Get project path
     eg. workspace/TestClass/src/TestClass.silecsdesign -> workspace/TestClass"""
    return os.path.dirname(os.path.dirname(filepath))


def get_schema_path(silecs_design_deploy_path):
    """Get schema for silecs design/deploy XML validation"""
    extension = get_extension(silecs_design_deploy_path)
    if extension == SILECSDESIGN:
        return DESIGN_SHEMA_PATH
    if extension == SILECSDEPLOY:
        return DEPLOY_SHEMA_PATH

    raise Exception("Error: Passed file for validation needs to have the extension '.silecsdesign' or '.silecsdeploy'")


def get_silecs_file_path_from_fesa(fesa_design_deploy_path):
    """Get schema for silecs design/deploy XML validation"""
    extension = get_extension(fesa_design_deploy_path)
    project_name = get_project_name(fesa_design_deploy_path)
    project_dir = get_project_dir(fesa_design_deploy_path)

    if extension == DESIGN:
        extension = SILECSDESIGN
    elif extension == DEPLOY:
        extension = SILECSDEPLOY
    else:
        raise Exception("Error: Passed FESA file needs to have the extension '.design' or '.deploy'")

    return os.path.join(project_dir, project_name + "." + extension)


def create_backup_file(filepath):
    backup_file = filepath + ".backup"
    with open(backup_file, 'w+'):
        shutil.copyfile(src=filepath, dst=backup_file)

##### Validate silecs design deploy ######

def validate(silecs_design_deploy_path, xsd_path):
    """Validate silecs design deploy"""
    try:
        print(f"validating {silecs_design_deploy_path}")
        xmlschema_doc = lxml.etree.parse(xsd_path)
        xmlschema = lxml.etree.XMLSchema(xmlschema_doc)
        xml_doc = lxml.etree.parse(silecs_design_deploy_path)
        result = xmlschema.validate(xml_doc)
        return result
    except OSError as e:
        print(e)
        return False
    except Exception as e:
        print(e)
        return False

# TODO: Implement jave-rules -  See here for details: https://gitlab.com/al.schwinn/silecs-cli/-/issues/3

def silecs_validate(file_paths):
    for path in file_paths:
        try:
            xsd_path = get_schema_path(path)
            val_result = validate(path, xsd_path)
            if val_result:
                print("File is valid")
            else:
                print("File is not valid. Check errors above.")
        except Exception as e:
            print(e)

##### Create silecs design/deploy file #####

def create_silecs_design(filepath, project_name):
    """"Create silecs design file"""
    date = datetime.datetime.today().strftime('%m/%d/%Y')
    design_template = f"""<?xml version="1.0" encoding="UTF-8"?>
<SILECS-Design silecs-version="{SILECS_VERSION}" created="{date}" updated="{date}"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:noNamespaceSchemaLocation="{DESIGN_SHEMA_PATH}">
    <Information>
        <Owner user-login="{USER}"/>
        <Editor user-login="{USER}"/>
    </Information>
    <SILECS-Class name="{project_name}" version="0.1.0" domain="OPERATIONAL" >
        <Acquisition-Block name="MyBlock" generateFesaProperty="true">
            <Acquisition-Register name="myRegister" generateFesaValueItem="true">
                <scalar format="int32"/>
            </Acquisition-Register>
        </Acquisition-Block>
    </SILECS-Class>
</SILECS-Design>"""

    if os.path.isfile(filepath):
        raise Exception(f"Error: There is already a .silecsdesign file available: {os.path.abspath(filepath)}")

    with open(filepath, "w") as fp:
        fp.write(design_template)

    print(f"\ncreated new silecsdesign: {filepath}")


def create_silecs_deploy(filepath, project_name):
    """"Create silecs deploy file"""
    date = datetime.datetime.today().strftime('%m/%d/%Y')
    deploy_template=f"""<?xml version="1.0" encoding="UTF-8"?>
<SILECS-Deploy silecs-version="{SILECS_VERSION}" created="{date}" updated="{date}" 
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:noNamespaceSchemaLocation="{DEPLOY_SHEMA_PATH}">
    <Information>
        <Owner user-login="{USER}"/>
        <Editor user-login="{USER}"/>
    </Information>
    <Deploy-Unit name="{project_name}" version="0.1.0"/>
    <SilecsDesign silecs-design-name="" silecs-design-version=""/>
        
    <Controller host-name="">
        <Siemens-PLC system="TIA-PORTAL" model="SIMATIC_S7-300" protocol="DEVICE_MODE" base-DB-number="1">
            <Device silecs-device-label="dev0" silecs-design-ref="" fesa-device-name="" fesa-fec-name=""/>
        </Siemens-PLC>
    </Controller>
</SILECS-Deploy>"""

    if os.path.isfile(filepath):
        raise Exception(f"Error: There is already a .silecsdeploy file available: {os.path.abspath(filepath)}")

    with open(filepath, "w") as fp:
        fp.write(deploy_template)

    print(f"\ncreated new silecsdeploy: {filepath}")


def create(filepath):
    """Create silecs design/deploy file"""
    project_name = get_project_name(filepath)
    extension = get_extension(filepath)

    if extension == SILECSDESIGN:
        create_silecs_design(filepath, project_name)
    elif extension == SILECSDEPLOY:
        create_silecs_deploy(filepath, project_name)


def silecs_create(file_paths):
    for path in file_paths:
        path = os.path.abspath(path)
        try:
            new_file_path = get_silecs_file_path_from_fesa(path)
            create(new_file_path)
            print("Created")
        except Exception as e:
            print(e)


##### Generation source code for silecs deploy path ######

def generate_silecsdeploy(silecs_deploy_path):
    """Generate source files from silecs deploy file"""
    project_name = get_project_name(silecs_deploy_path)
    project_path = get_project_path(silecs_deploy_path)
    workspace_path = os.path.abspath(os.path.join(project_path, os.pardir))
    fesa_deploy_file = os.path.abspath(os.path.join(project_path, project_name + ".depoy"))
    SILECS_DEPLOY_VERSION="unused_parameter"

    print(f"generating code for {project_name} ...\n")
    print(fesa_deploy_file)
    if os.path.isfile(fesa_deploy_file):
        info = f"Please note that old, obsolete xml-elements are not deleted automatically! Sometimes this can lead to an invalid FESA document! \n" \
                f"If this is the case, please remove the obsolete elements by hand. \n" \
                f"Creating a backup of the existing FESA file at: {fesa_deploy_file}.backup \n"

        print(info)
        create_backup_file(fesa_deploy_file)


    genparam.genParam(workspace_path, project_name, SILECS_DEPLOY_VERSION, SILECS_VERSION)
    genplcsrc.genPlcSrc(workspace_path, project_name, SILECS_VERSION)
    genduwrapper.genDuWrapper(workspace_path, project_name, SILECS_DEPLOY_VERSION)
    fillFESADeployUnit.fillDeployUnit(workspace_path, project_name, FESA_XSD, FESA_VERSION)

    info = "Please note that you need to add the dependency to the silecs library to your Makefile. \n" \
    	    "Here an example: https://www-acc.gsi.de/wiki/Frontend/Silecs_1_0_0_CodeSnippets#Dependency_in_Makefile.specific !\n"
    print(info)
    print("Code generation finished")


def generate_silecsdesign(silecs_design_path):
    """Generate source files from silecs design file"""
    project_name = get_project_name(silecs_design_path)
    project_path = get_project_path(silecs_design_path)
    workspace_path = os.path.abspath(os.path.join(project_path, os.pardir))
    fesa_design_file = os.path.abspath(os.path.join(project_path, project_name + ".design"))

    print(f"generating code for {project_name} ...\n")
    print(fesa_design_file)
    if os.path.isfile(fesa_design_file):
        info = f"Please note that old, obsolete xml-elements are not deleted automatically! Sometimes this can lead to an invalid FESA document! \n" \
                f"If this is the case, please remove the obsolete elements by hand. \n" \
                f"Creating a backup of the existing FESA file at: {fesa_design_file}.backup \n"
        print(info)
        create_backup_file(fesa_design_file)

    info = "Please not that you need to add the dependency to the silecs library to your Makefile. \n" \
    	    "Here an example: https://www-acc.gsi.de/wiki/Frontend/Silecs_1_0_0_CodeSnippets#Dependency_in_Makefile.specific !\n"
    print(info)

    generateFesaDesign.fillDesignFile(FESA_VERSION, project_name, workspace_path, FESA_XSD)
    generateSourceCode.genCppFiles(project_name, workspace_path, silecs_design_path)

    print("Code generation finished")

def generate(silecs_design_deploy_path):
    """Generate source files"""
    extension = get_extension(silecs_design_deploy_path)
    silecs_design_deploy_path = os.path.abspath(silecs_design_deploy_path)

    if extension == SILECSDESIGN:
        generate_silecsdesign(silecs_design_deploy_path)
    elif extension == SILECSDEPLOY:
        generate_silecsdeploy(silecs_design_deploy_path)


def silecs_generate(file_paths):
    for path in file_paths:
        try:
            xsd_path = get_schema_path(path)
            val_result = validate(path, xsd_path)
            if val_result:
                print("File is valid")
            else:
                print("File is not valid. Check errors above.")
                continue
            generate(path)
        except Exception as e:
            print(e)


##### Open diagnostic ######
def silecs_diagnostic_tool(filepath):
    filepath = os.path.abspath(filepath)
    print(filepath)
    extension = get_extension(filepath)
    if extension != SILECSDEPLOY:
        print("Error: diag tool only can be opened for a -silecsdeploy file")
        sys.exit()

    silecs_diag_tool=f"{SILECS_BASE}/silecs-diagnostic-cpp/bin/x86_64/silecs-diagnostic"
    subprocess.call([silecs_diag_tool, "-d", filepath])


##### Release parameter file ######

def silecs_release(file_paths):
    # TODO -  See here for details: https://gitlab.com/al.schwinn/silecs-cli/-/issues/5
    print("Not implemented yet .. please copy the param file by hand for now")


##### Migrate file to new silecs version ######

def silecs_migrate(file_paths):
# TODO -  See here for details: https://gitlab.com/al.schwinn/silecs-cli/-/issues/7
    print("Not implemented yet .. please update/migrate by hand for now")

##### Parse arguments ######

def _parse_arguments():
    parser = SilecsArgumentParser(
        description="Script in order to model PLC devices in FESA by using the Silecs framework",
        add_help=False
    )

    parser.add_argument(
        '-h', 
        '--help',
        action='help',
        help='Show this help message and exit.'
    )

    parser.add_argument(
        "-c",
        "--create",
        nargs="+",
        metavar="file",
        help="create new .silecsdesign/.silecsdeploy file for the specified FESA .design/.deploy file"
    )

    parser.add_argument(
        "-v",
        "--validate",
        nargs="+",
        metavar="file",
        help="validate silecs xml file"
    )

    parser.add_argument(
        "-g",
        "--generate",
        nargs="+",
        metavar="file",
        help="generate source code for silecs xml file"
    )

    parser.add_argument(
        "-d",
        "--diagnostic",
        nargs=1,
        metavar="file",
        help="start diagnostic tool for silecs xml file"
    )

    parser.add_argument(
        "-r",
        "--release",
        metavar="file",
        help="release parameter file fo FEC (TODO)"
    )

    parser.add_argument(
        "-m",
        "--migrate",
        metavar="file",
        help="migrate silecs design/deploy to new silecs version (TODO)"
    )

    return parser


##### Entry point ######

def run():
    parser = _parse_arguments()
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)

    try:
        options = parser.parse_args()
    except Exception:
        sys.exit(0)

    if options.create:
        silecs_create(options.create)
    if options.validate:
        silecs_validate(options.validate)
    if options.generate:
        silecs_generate(options.generate)
    if options.diagnostic:
        filepath = options.diagnostic[0]
        silecs_diagnostic_tool(filepath)
    if options.release:
        silecs_release(options.release)
    if options.migrate:
        silecs_migrate(options.migrate)

if __name__ == "__main__":
    run()
