import os
import sys
import unittest

sys.path.append('../')
from silecs_cli import silecs

BASE_FOR_BACKUP = os.path.join("tests", "examples", "SilecsTestClass.design")
BACKUP = BASE_FOR_BACKUP + ".backup"

DESIGN_PATH = os.path.join("tests", "examples", "SilecsTestClass2.design")
DEPLOY_PATH = os.path.join("tests", "examples", "SilecsTestClass_2_DU.deploy")

SILECS_DESIGN_PATH = DESIGN_PATH.split(".")[0] + ".silecsdesign"
SILECS_DEPLOY_PATH = DEPLOY_PATH.split(".")[0] + ".silecsdeploy"


class SilecsTest(unittest.TestCase):
    """Class to test silecs.py script"""
    @classmethod
    def tearDownClass(cls):
        """Remove created files during tests"""
        if os.path.isfile(BACKUP):
            os.remove(BACKUP)

        if os.path.isfile(SILECS_DESIGN_PATH):
            os.remove(SILECS_DESIGN_PATH)

        if os.path.isfile(SILECS_DEPLOY_PATH):
            os.remove(SILECS_DEPLOY_PATH)

    def test_get_extension(self):
        """Test getting extension from file"""
        filenames_extensions = [("test1.design", "design"), ("test2.deploy", "deploy")]
        for filename, expected in filenames_extensions:
            with self.subTest():
                self.assertEqual(silecs.get_extension(filename), expected)

    def test_get_project_name(self):
        """Test getting extension from filepath"""
        filenames_project_names = [("test1.design", "test1"), ("test2.deploy", "test2")]
        for filename, project_name in filenames_project_names:
            with self.subTest():
                self.assertEqual(silecs.get_project_name(filename), project_name)

    def test_get_project_path(self):
        """Test getting project path from filepath"""
        filenames_project_paths = [("workspace/test1/src/test1.design", "workspace/test1"),
                    ("workspace/test2/src/test2.deploy", "workspace/test2")]
        for filename, project_path in filenames_project_paths:
            with self.subTest():
                self.assertEqual(silecs.get_project_path(filename), project_path)

    def test_get_project_dir(self):
        """Test getting project directory from filepath"""
        filenames_dirs = [("workspace/test1/src/test1.design", "workspace/test1/src"),
                    ("workspace/test2/src/test2.deploy", "workspace/test2/src")]
        for filename, directory in filenames_dirs:
            with self.subTest():
                self.assertEqual(silecs.get_project_dir(filename), directory)

    def test_get_schema_path(self):
        """Test getting xml schema path for silecs design deploy files"""
        filenames_schema_paths = [("/home/test/example/test1.silecsdesign",
                         "../silecs-model/src/xml/DesignSchema.xsd"),
                            ("/home/test/example/test2.silecsdeploy",
                             "../silecs-model/src/xml/DeploySchema.xsd")]
        for filename, schema_path in filenames_schema_paths:
            with self.subTest():
                self.assertEqual(silecs.get_schema_path(filename), schema_path)

    def test_get_schema_path_invalid(self):
        """Test getting xml schema path for invalid path (passed file path
            must have the extension '.silecsdesign' or '.silecsdeploy')"""
        try:
            silecs.get_schema_path("/home/test/example/test1.design")
        except Exception as e:
            self.assertEqual("Error: Passed file for validation needs to have the extension '.silecsdesign' or '.silecsdeploy'", str(e))

    def test_get_silecs_file_path_from_fesa(self):
        """Test getting xml schema path for silecs design deploy files"""
        self.assertEqual(silecs.get_silecs_file_path_from_fesa("../test1.design"), os.path.join("..", "test1.silecsdesign"))

    def test_get_silecs_file_path_from_fesa_invalid(self):
        """Test getting xml schema path for invalid file (passed file path
            must have the extension '.design' or '.deploy')
        """
        try:
            silecs.get_silecs_file_path_from_fesa("../test1.silecsdesign")
        except Exception as e:
            self.assertEqual("Error: Passed FESA file needs to have the extension '.design' or '.deploy'", str(e))

    def test_create_backup_file(self):
        """Test creating backup file"""
        silecs.create_backup_file(BASE_FOR_BACKUP)
        self.assertNotEqual(os.stat(BACKUP).st_size,  0)

    def test_silecs_validate(self):
        """Test validating silecs design/deploy files"""
        silecs_design_deploy_paths = [os.path.join("tests", "examples", "SilecsTestClass.silecsdesign"), \
                                 os.path.join("tests", "examples", "SilecsTestClassDU.silecsdeploy")]
        for path in silecs_design_deploy_paths:
            xsd_path = silecs.get_schema_path(path)
            self.assertEqual(silecs.validate(path, xsd_path), True)

    def test_silecs_validate_invalid(self):
        """Test validating invalid silecs file"""
        invalid_silecs_deploy_path = os.path.join("tests", "examples", "SilecsTestClassDU_invalid.silecsdeploy")
        xsd_path = silecs.get_schema_path(invalid_silecs_deploy_path)
        self.assertEqual(silecs.validate(invalid_silecs_deploy_path, xsd_path), False)

    def test_create_silecs_design(self):
        """Test creating silecs design files from design file"""
        path = os.path.abspath(DESIGN_PATH)

        new_file_path = silecs.get_silecs_file_path_from_fesa(path)
        silecs.create(new_file_path)
        print(new_file_path)
        self.assertEqual(os.path.exists(new_file_path), True)
        self.assertNotEqual(os.stat(new_file_path).st_size,  0)

        try:
            silecs.create(new_file_path)
        except Exception as e:
            self.assertIn("There is already a .silecsdesign file available:", str(e))

    def test_create_silecs_deploy(self):
        """Test creating silecs deploy files from deploy file"""
        path = os.path.abspath(DEPLOY_PATH)

        new_file_path = silecs.get_silecs_file_path_from_fesa(path)
        silecs.create(new_file_path)
        print(new_file_path)
        self.assertEqual(os.path.exists(new_file_path), True)
        self.assertNotEqual(os.stat(new_file_path).st_size,  0)

        try:
            silecs.create(new_file_path)
        except Exception as e:
            print(str(e))
            self.assertIn("There is already a .silecsdeploy file available:", str(e))


if __name__ == "_main__":
    unittest.main()
